using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Drawing2D;
using System.Reflection;
using System.Xml.Serialization;
using System.IO;

namespace RakeTendril
{
    public partial class MainForm : Form
    {
        private List<PlacedTendril> tendrils = new List<PlacedTendril>();
        private List<FieldInfo> parameters = new List<FieldInfo>();
        private List<float> mins = new List<float>();
        private List<float> maxs = new List<float>();
        private bool isInitialized = false;
        internal const int bitmapWidth = 640;
        internal const int bitmapHeight = 480;
        private const int framesPerSecond = 30;
        internal Bitmap bitmap = new Bitmap(bitmapWidth, bitmapHeight); // only access with lock(bitmap)
        private Graphics graphics;
        private DisplayForm mainDisplayForm;
        private DisplayForm controlDisplayForm;
        private Font font = new Font("Arial Narrow", 8.0f);
        private List<Parameter> presets = new List<Parameter>();
        private string[] names = new string[0];
        private double transitionTime = 0.0;
        private double transitionPhase = 1.0;
        private double swayPhase = 0.0;
        private Parameter parameterA = new Parameter();
        private Parameter parameterB = new Parameter();
        private Communicator communicator;
        private DateTime dateTime = DateTime.Now;
        private int fpsCount = 0;
        private double fpsElapsed = 0.0;

        public MainForm()
        {
            graphics = Graphics.FromImage(bitmap);
            graphics.SmoothingMode = SmoothingMode.HighQuality;
            InitializeComponent();
            generateTendrils();

            foreach (FieldInfo fi in typeof(Parameter).GetFields())
            {
                if(fi.FieldType != typeof(float))
                {
                    continue;
                }
                object[] atts = fi.GetCustomAttributes(typeof(ParameterAttribute), false);
                if (atts == null || atts.Length == 0 || !(atts[0] is ParameterAttribute))
                {
                    continue;
                }
                ParameterAttribute pa = (ParameterAttribute)atts[0];
                if (pa == null)
                {
                    continue;
                }

                parameters.Add(fi);
                mins.Add(pa.Min);
                maxs.Add(pa.Max);
            }

            int i = 0;
            foreach (Control c in Controls)
            {
                if (i >= parameters.Count)
                {
                    break;
                }
                if (!(c is TrackBar))
                {
                    continue;
                }
                TrackBar t = (TrackBar)c;
                if (t == null)
                {
                    continue;
                }
                t.BackColor = Color.FromName(((ParameterAttribute)parameters[i].GetCustomAttributes(typeof(ParameterAttribute), false)[0]).Col);
                i++;
            }

            updateTrackBars();

            loadPresets();

            isInitialized = true;
            timer1.Interval = 1000 / framesPerSecond;
            timer1.Enabled = true;

            controlDisplayForm = new DisplayForm(this, "Control Display");
            controlDisplayForm.Show();
        }

        private void updateTrackBars()
        {
            int i = 0;
            foreach (Control c in Controls)
            {
                if (i >= parameters.Count)
                {
                    break;
                }
                if (!(c is TrackBar))
                {
                    continue;
                }
                TrackBar t = (TrackBar)c;
                if (t == null)
                {
                    continue;
                }
                t.Value = (int)(1000.0 * ((float)parameters[i].GetValue(Tendril.Param) - mins[i]) / (maxs[i] - mins[i]));
                i++;
            }
        }

        private void generateTendrils()
        {
            tendrils.Clear();

            Random rootRandom = new Random(23456);
            float countF = -1.0f+ 40.0f*(float)Math.Pow(Tendril.Param.Count, 2.0); // extend the scale at the lower end
            int count = (int)countF;
            float countFrac = countF - count;
            float remainder = 1.0f;
            const float deadZone = 0.02f;
            if (countFrac < deadZone)
            {
            }
            else if (countFrac > 1.0f - deadZone)
            {
                count++;
            }
            else
            {
                count++;
                float off = (countFrac - deadZone) / (1.0f - 2.0f * deadZone);
                //float ease = (3.0f  - 2.0f * off)*off*off;
                remainder = 0.00001f + off;
            }
            for (int i = 0; i < count; i++)
            {
                PlacedTendril pt = new PlacedTendril();
                tendrils.Add(pt);
                if (count <= 1)
                {
                    pt.Scale = remainder; // fractional part to have a smooth animation of the count
                    pt.ZPosition = 0.0f;
                }
                else if (count == 2)
                {
                    if (i == 0)
                    {
                        pt.Scale = 1.0f;
                        pt.ZPosition = 0.0f;
                    }
                    else
                    {
                        pt.Scale = remainder; // fractional part to have a smooth animation of the count
                        pt.ZPosition = 1.0f;
                    }
                }
                else if (i == count - 1)
                {
                    pt.Scale = remainder; // fractional part to have a smooth animation of the count
                    pt.ZPosition = 1.0f;
                }
                else
                {
                    pt.Scale = 1.0f;
                    pt.ZPosition = i / (count - 2.0f + remainder);
                }
                int r0 = rootRandom.Next(1000);
                //bool b00 = (r1 < Tendril.Param.LeavsPct1 * 999);
                bool b01 = (r0 < (Tendril.Param.BlssmPct0) * 999);
                pt.IsBlossom = b01;
                //pt.IsLeaf = (b00&&!b01);
                double px = rootRandom.NextDouble() + Tendril.Param.WShuffle;
                if (px >= 1.0)
                {
                    px -= 1.0;
                }
                pt.Location = new PointF((float)px, 0.0f);
                bool curlsLeft = (rootRandom.Next(2) == 0);  //(i % 2 == 0)
                pt.CurlsLeft = curlsLeft;
                double swayAngle = 0.5*Tendril.Param.SwayAmnt * Math.Sin(2.0 * Math.PI * (swayPhase+Tendril.Param.SwayAsync*rootRandom.NextDouble()));
                pt.Angle = (float)(swayAngle + Tendril.Param.RootAng + 2.0 * Tendril.Param.RootAngVar * ((i % 2 == 0)?1.0:-1.0)*(0.2+0.2*rootRandom.NextDouble()));
                float spiralWindOffset = (float)(Tendril.Param.SprWndVar*((i % 2 == 0)?1.0:-1.0)*(0.2+0.2*rootRandom.NextDouble()));
                pt.SpiralWindOffset = spiralWindOffset;

                Random branchRandom1a = new Random(rootRandom.Next(10000));
                double gen1F = Tendril.Param.BranchCnt1 + 2.0 * branchRandom1a.NextDouble() - 1.0;
                int gen1 = (int)gen1F;
                double gen1Frac = gen1F - gen1;
                double[] pos1 = new double[gen1];
                double sumPos1 = branchRandom1a.NextDouble() + 0.1; // gap below and prevent division by zero
                for (int g1 = 0; g1 < gen1; g1++)
                {
                    pos1[g1] = sumPos1;
                    double rnd1 = 0.2+0.8*branchRandom1a.NextDouble();
                    if (g1 == gen1 - 1)
                    {
                        sumPos1 += gen1Frac * rnd1;
                    }
                    else
                    {
                        sumPos1 += rnd1;
                    }
                }
                for (int g1 = 0; g1 < gen1; g1++)
                {
                    pos1[g1] /= sumPos1;
                }
                Random branchRandom1 = new Random(rootRandom.Next(10000));
                for (int g1 = 0; g1 < gen1; g1++)
                {
                    Tendril t1 = new Tendril();
                    pt.Children.Add(t1);
                    int r1 = branchRandom1.Next(1000);
                    bool b10 = (r1 > Tendril.Param.LeavsPct1 * 999);
                    bool b11 = (r1 < (Tendril.Param.LeavsPct1 + Tendril.Param.BlssmPct1) * 999);
                    t1.IsBlossom = (b11 && b10);
                    t1.IsLeaf = (b11 && !b10);
                    t1.Position = (float)pos1[g1];// 0.1f + 0.7f * (float)branchRandom1.NextDouble();
                    bool curlsLeft1 = (g1%2==0);//(branchRandom1.Next(2) == 0);
                    t1.CurlsLeft = curlsLeft1;
                    double swayAngle1 = 0.7*Tendril.Param.SwayAmnt * Math.Sin(2.0 * Math.PI * (swayPhase + Tendril.Param.SwayAsync * branchRandom1.NextDouble()));
                    t1.Angle = (float)(-swayAngle1 + Tendril.Param.BranchAng * (curlsLeft1 ? 1.0 : -1.0));
                    t1.SpiralWindOffset = spiralWindOffset;

                    Random branchRandom2a = new Random(branchRandom1.Next(10000));
                    double gen2F = Tendril.Param.BranchCnt2 + 2.0 * branchRandom2a.NextDouble() - 1.0;
                    int gen2 = (int)gen2F;
                    double gen2Frac = gen2F - gen2;
                    double[] pos2 = new double[gen2];
                    double sumPos2 = branchRandom2a.NextDouble() + 0.1; // gap below and prevent division by zero
                    for (int g2 = 0; g2 < gen2; g2++)
                    {
                        pos2[g2] = sumPos2;
                        double rnd2 = 0.2 + 0.8 * branchRandom2a.NextDouble();
                        if (g2 == gen2 - 1)
                        {
                            sumPos2 += gen2Frac * rnd2;
                        }
                        else
                        {
                            sumPos2 += rnd2;
                        }
                    }
                    for (int g2 = 0; g2 < gen2; g2++)
                    {
                        pos2[g2] /= sumPos2;
                    }
                    Random branchRandom2 = new Random(branchRandom1.Next(10000));
                    for (int g2 = 0; g2 < gen2; g2++)
                    {
                        Tendril t2 = new Tendril();
                        t1.Children.Add(t2);
                        int r2 = branchRandom2.Next(1000);
                        bool b20 = (r2 < Tendril.Param.LeavsPct2 * 999);
                        //bool b11 = (r1 < (Tendril.Param.LeavsPct1 + Tendril.Param.BlssmPct1) * 999);
                        //t1.IsBlossom = b11;
                        t2.IsLeaf = b20; // (b10 && !b11);
                        t2.Position = (float)pos2[g2];// 0.1f + 0.7f * (float)branchRandom2.NextDouble();
                        bool curlsLeft2 = (g2 % 2 == 0); // (branchRandom2.Next(2) == 0);
                        t2.CurlsLeft = curlsLeft2;
                        double swayAngle2 = Tendril.Param.SwayAmnt * Math.Sin(2.0 * Math.PI * (swayPhase + Tendril.Param.SwayAsync * branchRandom2.NextDouble()));
                        t2.Angle = (float)(-swayAngle2 + Tendril.Param.BranchAng * (curlsLeft2 ? 1.0 : -1.0));
                        t2.SpiralWindOffset = spiralWindOffset;
                    }
                }
            }
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            switch(e.KeyCode)
            {
                case Keys.Space:
                    e.Handled = true;
                    if (listBoxPresets.SelectedIndex < listBoxPresets.Items.Count - 1)
                    {
                        listBoxPresets.SelectedIndex++;
                    }
                    break;
                default:
                    break;
            }

            base.OnKeyDown(e);
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            int i = 0;
            foreach (Control c in Controls)
            {
                if (i >= parameters.Count)
                {
                    break;
                }
                if (!(c is TrackBar))
                {
                    continue;
                }
                TrackBar t = (TrackBar)c;
                if (t == null)
                {
                    continue;
                }
                e.Graphics.DrawString(parameters[i].Name, font, Brushes.Black, t.Location.X, t.Location.Y -17);
                i++;
            }
        }

        private void trackBar_ValueChanged(object sender, EventArgs e)
        {
            if (!isInitialized)
            {
                return;
            }

            int i = 0;
            foreach (Control c in Controls)
            {
                if (i >= parameters.Count)
                {
                    break;
                }
                if(!(c is TrackBar))
                {
                    continue;
                }
                TrackBar t = (TrackBar)c;
                if (t == null)
                {
                    continue;
                }
                parameters[i].SetValue(Tendril.Param, mins[i] + (maxs[i] - mins[i]) * 0.001f * t.Value);
                i++;
            }
        }

        private void buttonShowTendrils_Click(object sender, EventArgs e)
        {
            if (mainDisplayForm == null || mainDisplayForm.IsDisposed)
            {
                mainDisplayForm = new DisplayForm(this, "Main Display");
            }

            if (!mainDisplayForm.Visible)
            {
                mainDisplayForm.Show();
            }
        }

        private Color fadeToBack(int alpha, float r, float g, float b, double blend)
        {
            double inv = 1.0-blend;
            return Color.FromArgb(alpha, (int)(inv * r + blend * Tendril.Param.BackR), (int)(inv * g + blend * Tendril.Param.BackG), (int)(inv * b + blend * Tendril.Param.BackB));
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            DateTime now = DateTime.Now;
            double elapsedTime = (now - dateTime).TotalSeconds;
            dateTime = now;

            fpsCount++;
            fpsElapsed += elapsedTime;
            if (fpsCount == 5)
            {
                Text = Math.Round(5/(fpsElapsed+0.0001), 1) + " fps";
                fpsCount = 0;
                fpsElapsed = 0.0;
            }

            swayPhase += elapsedTime / (Tendril.Param.SwayTime+0.001);
            swayPhase -= Math.Floor(swayPhase); // keep the fractional part for more exact frequency

            if (transitionPhase < 1.0)
            {
                transitionPhase = Math.Min(1.0, transitionPhase + elapsedTime / (transitionTime+0.001));
                for (int i = 0; i < parameters.Count; i++)
                {
                    float a = (float)parameters[i].GetValue(parameterA);
                    float b = (float)parameters[i].GetValue(parameterB);
                    // ease in/out
                    double t = transitionPhase * transitionPhase * (3.0 - 2.0 * transitionPhase);
                    parameters[i].SetValue(Tendril.Param, (float)((1.0-t)*a + t*b));    
                }
                updateTrackBars();
            }

            if (transitionPhase > 0.99)
            {
                BackColor = Color.LightGray;
            }

            generateTendrils();

            lock (bitmap)
            {
                graphics.Clear(Color.FromArgb((int)Tendril.Param.BackR, (int)Tendril.Param.BackG, (int)Tendril.Param.BackB));
                graphics.Transform = new Matrix();
                graphics.TranslateTransform(0, bitmapHeight - 1);
                graphics.ScaleTransform(0.0001f + bitmapWidth, 0.0001f + bitmapWidth);
                //render backwards: the frontmost last
                for (int i = tendrils.Count-1; i >= 0; i--)
                {
                    PlacedTendril t = tendrils[i];
                    Matrix m = graphics.Transform;
                    float h = 1.0f - (float)(0.01 * (double)numericUpDownClampBottom.Value);
                    graphics.TranslateTransform(0.5f-0.5f*h, bitmapHeight/(0.01f+bitmapWidth)*(h-1.0f));
                    graphics.ScaleTransform(h, h);
                    graphics.TranslateTransform((t.Location.X - 0.5f) * Tendril.Param.Width, t.Location.Y * Tendril.Param.Size);
                    graphics.TranslateTransform(Tendril.Param.PosX, -Tendril.Param.PosY);
                    float size = (float)(0.0001f + t.Scale*(1.0f + Tendril.Param.SizeVar * (0.5 - t.ZPosition)) * Tendril.Param.Size);
                    graphics.ScaleTransform(size, size);
                    graphics.RotateTransform(-90.0f + t.Angle);
                    double blend = Tendril.Param.FadeToBck * Math.Sqrt(t.ZPosition);
                    int alpha = (int)(255.9 * (1.0 - Tendril.Param.Transpncy));
                    t.Paint(graphics, Tendril.Param.StemWidth, 1.0f,
                        new SolidBrush(fadeToBack(alpha, Tendril.Param.TendrilR, Tendril.Param.TendrilG, Tendril.Param.TendrilB, blend)),
                        new SolidBrush(fadeToBack(alpha, Tendril.Param.BlssmFillR, Tendril.Param.BlssmFillG, Tendril.Param.BlssmFillB, blend)),
                        fadeToBack(alpha, Tendril.Param.BlssmPenR, Tendril.Param.BlssmPenG, Tendril.Param.BlssmPenB, blend));
                    graphics.Transform = m;
                }
                if (communicator != null)
                {
                    communicator.NewImageWaiting = true;
                }
            }

            controlDisplayForm.Invalidate();

            if (mainDisplayForm != null && !mainDisplayForm.IsDisposed)
            {
                mainDisplayForm.Invalidate();
            }

            if (communicator == null)
            {
                labelNumFrames.Text = "no connection";
            }
            else
            {
                labelNumFrames.Text = System.Threading.Interlocked.Read(ref communicator.FramesSent).ToString();
            }
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            if (communicator != null)
            {
                communicator.Stop();
                while (!communicator.Stopped)
                {
                    System.Threading.Thread.Sleep(100);
                }
            }

            base.OnClosing(e);
        }

        private void savePresets()
        {
            try
            {
                XmlSerializer serializer = new XmlSerializer(typeof(List<Parameter>));
                FileStream file = new FileStream("presets" + DateTime.Now.Ticks + ".xml", FileMode.CreateNew);
                serializer.Serialize(file, presets);
                file.Close();
            }
            catch (Exception e)
            {
                MessageBox.Show("Couldn't save presets:\n" + e.Message);
            }
        }

        private void loadPresets()
        {
            List<Parameter> paras;
            try
            {
                string[] files = Directory.GetFiles(Directory.GetCurrentDirectory(), "presets*.xml", SearchOption.TopDirectoryOnly);
                Array.Sort(files);
                XmlSerializer serializer = new XmlSerializer(typeof(List<Parameter>));
                FileStream file = new FileStream(files[files.Length-1], FileMode.Open);
                paras = (List<Parameter>)serializer.Deserialize(file);
                file.Close();
                presets = paras;
                populatePresetList();
            }
            catch (Exception e)
            {
                MessageBox.Show("Couldn't load presets:\n" + e.Message);
            }
        }

        private void populatePresetList()
        {
            names = new string[presets.Count];
            for (int i = 0; i < presets.Count; i++)
            {
                names[i] = presets[i].Name;
            }
            Array.Sort(names);
            listBoxPresets.Items.Clear();
            listBoxPresets.Items.AddRange(names);
        }

        private void buttonAdd_Click(object sender, EventArgs e)
        {
            if (textBoxNewName.Text.Length == 0)
            {
                MessageBox.Show("Enter a name");
                return;
            }
            bool found = false;
            foreach (Parameter r in presets)
            {
                if(r.Name.Equals(textBoxNewName.Text))
                {
                    found = true;
                    break;
                }
            }
            if(found)
            {
                MessageBox.Show("This name is already taken.");
                return;
            }

            Parameter p = Tendril.Param.Clone();
            p.Name = textBoxNewName.Text;
            presets.Add(p);
            populatePresetList();
            savePresets();
        }

        private void listBoxPresets_SelectedIndexChanged(object sender, EventArgs e)
        {
            if(listBoxPresets.SelectedIndex < 0)
            {
                return;
            }
            string name = names[listBoxPresets.SelectedIndex];
            foreach (Parameter r in presets)
            {
                if (r.Name.Equals(name))
                {
                    parameterA = Tendril.Param.Clone();
                    parameterB = r.Clone();
                    transitionPhase = 0.0;
                    BackColor = Color.LightPink;
                    break;
                }
            }
            textBoxNewName.Text = name;
        }

        private void radioButtonTransitionTime_Click(object sender, EventArgs e)
        {
            transitionTime = 0.0;
            if (radioButton2.Checked)
            {
                transitionTime = 0.2;
            }
            else if (radioButton3.Checked)
            {
                transitionTime = 0.5;
            }
            else if (radioButton4.Checked)
            {
                transitionTime = 1.0;
            }
            else if (radioButton5.Checked)
            {
                transitionTime = 2.0;
            }
            else if (radioButton6.Checked)
            {
                transitionTime = 5.0;
            }
            else if (radioButton7.Checked)
            {
                transitionTime = 10.0;
            }
            else if (radioButton8.Checked)
            {
                transitionTime = 20.0;
            }
            else if (radioButton9.Checked)
            {
                transitionTime = 50.0;
            }
            else if (radioButton10.Checked)
            {
                transitionTime = 100.0;
            }
            else if (radioButton11.Checked)
            {
                transitionTime = 200.0;
            }
            else if (radioButton12.Checked)
            {
                transitionTime = 500.0;
            }
            else if (radioButton13.Checked)
            {
                transitionTime = 1000.0;
            }
        }

        private void buttonRename_Click(object sender, EventArgs e)
        {
            if (listBoxPresets.SelectedIndex < 0)
            {
                MessageBox.Show("No preset selected");
                return;
            }
            if (textBoxNewName.Text.Length == 0)
            {
                MessageBox.Show("Enter a name");
                return;
            }
            bool found = false;         
            foreach (Parameter r in presets)
            {
                if (r.Name.Equals(textBoxNewName.Text))
                {
                    found = true;
                    break;
                }
            }
            if (found)
            {
                MessageBox.Show("This name is already taken.");
                return;
            }

            foreach (Parameter r in presets)
            {
                if (r.Name.Equals(listBoxPresets.Items[listBoxPresets.SelectedIndex]))
                {
                    r.Name = textBoxNewName.Text;
                    break;
                }
            }
            //alphabetic order may have changed
            populatePresetList();
            //select the right one
            for (int i = 0; i < presets.Count; i++)
            {
                if (names[i].Equals(textBoxNewName.Text))
                {
                    listBoxPresets.SelectedIndex = i;
                    break;
                }
            }

            savePresets();
        }

        private void buttonRemove_Click(object sender, EventArgs e)
        {
            if (listBoxPresets.SelectedIndex < 0)
            {
                MessageBox.Show("No preset selected");
                return;
            }
            bool found = false;
            int i = 0;
            for (; i < presets.Count; i++)
            {
                if (presets[i].Name.Equals(listBoxPresets.Items[listBoxPresets.SelectedIndex]))
                {
                    found = true;
                    break;
                }                
            }
            if (!found)
            {
                return;
            }
            presets.RemoveAt(i);
            populatePresetList();

            savePresets();
        }

        private void buttonConnect_Click(object sender, EventArgs e)
        {
            if (communicator == null)
            {
                try
                {
                    communicator = new Communicator(textBoxAddr.Text, Int32.Parse(textBoxPort.Text), this);
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Could not start communication:\n" + ex.Message);
                }
            }
            else
            {
                MessageBox.Show("There's still a connection (partially?) running.");
            }
        }

        private void buttonDisconnect_Click(object sender, EventArgs e)
        {
            if (communicator != null)
            {
                communicator.Stop();
                communicator = null;
            }
        }

        private void buttonHideTendrils_Click(object sender, EventArgs e)
        {
            if (mainDisplayForm != null && ! mainDisplayForm.IsDisposed)
            {
                mainDisplayForm.Close();
            }
        }

        private void buttonWindowFull_Click(object sender, EventArgs e)
        {
            if (mainDisplayForm != null && !mainDisplayForm.IsDisposed)
            {
                if (mainDisplayForm.WindowState == FormWindowState.Maximized)
                {
                    mainDisplayForm.FormBorderStyle = FormBorderStyle.Sizable;
                    mainDisplayForm.WindowState = FormWindowState.Normal;
                }
                else
                {
                    mainDisplayForm.FormBorderStyle = FormBorderStyle.None;
                    mainDisplayForm.WindowState = FormWindowState.Maximized;
                }
            }
        }

        private void buttonReopenControl_Click(object sender, EventArgs e)
        {
            if (controlDisplayForm == null || controlDisplayForm.IsDisposed)
            {
                controlDisplayForm = new DisplayForm(this, "Control Display");
            }

            if (!controlDisplayForm.Visible)
            {
                controlDisplayForm.Show();
            }
        }
    }

    [AttributeUsage(AttributeTargets.Field, AllowMultiple = false, Inherited = false)]
    public class ParameterAttribute : Attribute
    {
        public float Min;
        public float Max;
        public string Col;
    }

    [Serializable]
    public class Parameter // a struct would be nicer, but we want member initialization
    {
        public string Name;

        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "LightBlue")]
        public float Bloom = 1.0f;
        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "LightBlue")]
        public float BlssmSide = 0.2f;
        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "LightBlue")]
        public float BlssmWdth = 0.2f;
        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "LightBlue")]
        public float BlssmSize = 0.3f;
        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "LightBlue")]
        public float BlssmPct1 = 0.0f;
        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "LightBlue")]
        public float BlssmPct0 = 0.0f;

        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "Coral")]
        public float LeavsSize = 0.5f;
        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "Coral")]
        public float LeavsPct2 = 0.0f;
        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "Coral")]
        public float LeavsPct1 = 0.0f;

        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "Pink")]
        public float SizeDrop = 0.6f;
        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "Pink")]
        public float StemScale = 0.9f;
        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "Pink")]
        public float SizeScale = 0.9f;
        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "Pink")]
        public float Bulge = 0.4f;
        [ParameterAttribute(Min = -10.0f, Max = 10.0f, Col = "Pink")]
        public float SprWndVar = 0.0f;
        [ParameterAttribute(Min = -7.0f, Max = 7.0f, Col = "Pink")]
        public float SpiralWind = 2.7f;
        [ParameterAttribute(Min = 0.1f, Max = 20.0f, Col = "Pink")]
        public float SpiralTip = 5.2f;
        [ParameterAttribute(Min = 0.0f, Max = 0.3f, Col = "Pink")]
        public float StemWidth = 0.1f;
        [ParameterAttribute(Min = 0.0f, Max = 10.0f, Col = "Pink")]
        public float BranchCnt2 = 5.0f;
        [ParameterAttribute(Min = 0.0f, Max = 10.0f, Col = "Pink")]
        public float BranchCnt1 = 5.0f;

        [ParameterAttribute(Min = 0.0f, Max = 255.0f, Col = "Gray")]
        public float BlssmPenB = 80.0f;
        [ParameterAttribute(Min = 0.0f, Max = 255.0f, Col = "Gray")]
        public float BlssmPenG = 30.0f;
        [ParameterAttribute(Min = 0.0f, Max = 255.0f, Col = "Gray")]
        public float BlssmPenR = 255.0f;
        [ParameterAttribute(Min = 0.0f, Max = 255.0f, Col = "Gray")]
        public float BlssmFillB = 100.0f;
        [ParameterAttribute(Min = 0.0f, Max = 255.0f, Col = "Gray")]
        public float BlssmFillG = 150.0f;
        [ParameterAttribute(Min = 0.0f, Max = 255.0f, Col = "Gray")]
        public float BlssmFillR = 255.0f;
        [ParameterAttribute(Min = 0.0f, Max = 255.0f, Col = "Gray")]
        public float TendrilB = 80.0f;
        [ParameterAttribute(Min = 0.0f, Max = 255.0f, Col = "Gray")]
        public float TendrilG = 180.0f;
        [ParameterAttribute(Min = 0.0f, Max = 255.0f, Col = "Gray")]
        public float TendrilR = 40.0f;

        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "Indigo")]
        public float FadeToBck = 0.2f;
        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "Indigo")]
        public float Transpncy = 0.1f;
        [ParameterAttribute(Min = 0.0f, Max = 255.0f, Col = "Indigo")]
        public float BackB = 255.0f;
        [ParameterAttribute(Min = 0.0f, Max = 255.0f, Col = "Indigo")]
        public float BackG = 180.0f;
        [ParameterAttribute(Min = 0.0f, Max = 255.0f, Col = "Indigo")]
        public float BackR = 150.0f;

        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "LightGreen")]
        public float SwayAsync = 0.2f;
        [ParameterAttribute(Min = 0.5f, Max = 20.0f, Col = "LightGreen")]
        public float SwayTime = 3.0f;
        [ParameterAttribute(Min = 0.0f, Max = 45.0f, Col = "LightGreen")]
        public float SwayAmnt = 5.0f;

        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "BlanchedAlmond")]
        public float WShuffle = 0.5f;
        [ParameterAttribute(Min = 0.0f, Max = 2.0f, Col = "BlanchedAlmond")]
        public float SizeVar = 0.2f;
        [ParameterAttribute(Min = -60.0f, Max = 60.0f, Col = "BlanchedAlmond")]
        public float BranchAng = 20.0f;
        [ParameterAttribute(Min = -60.0f, Max = 60.0f, Col = "BlanchedAlmond")]
        public float RootAngVar = 20.0f;
        [ParameterAttribute(Min = -45.0f, Max = 45.0f, Col = "BlanchedAlmond")]
        public float RootAng = 10.0f;

        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "Red")]
        public float Count = 0.225f;
        [ParameterAttribute(Min = 0.0f, Max = 2.0f, Col = "Red")]
        public float Width = 0.3f;
        [ParameterAttribute(Min = 0.0f, Max = 1.0f, Col = "Red")]
        public float Size = 0.4f;
        [ParameterAttribute(Min = -0.5f, Max = 1.0f, Col = "Red")]
        public float PosY = 0.0f;
        [ParameterAttribute(Min = -0.5f, Max = 1.5f, Col = "Red")]
        public float PosX = 0.5f;

        public Parameter Clone()
        {
            return (Parameter)MemberwiseClone();
        }
    }

    class PlacedTendril : Tendril
    {
        public PointF Location = new PointF(0.0f, 0.0f);
        public float ZPosition = 0.0f;
        public float Scale = 1.0f;
    }

    class Tendril
    {
        public static Parameter Param = new Parameter(); 
        public float Position = 0.0f;
        public float Angle = 0.0f;
        public float SpiralWindOffset = 0.0f;
        public bool CurlsLeft = false;
        public bool IsBlossom = false;
        public bool IsLeaf = false;
        public List<Tendril> Children = new List<Tendril>();
        private static readonly GraphicsPath leafPath;

        static Tendril()
        {
            leafPath = new GraphicsPath();
            leafPath.AddBezier(0.0f, 0.0f, 0.19f, -0.06f, 0.38f, 0.62f, 1.0f, 0.38f);
            leafPath.AddBezier(1.0f, 0.38f, 0.28f, 1.02f, 0.18f, 0.45f, 0.16f, 0.22f);
            leafPath.AddBezier(0.16f, 0.22f, 0.28f, 0.45f, 0.38f, 0.55f, 0.62f, 0.5f);
            leafPath.AddBezier(0.62f, 0.5f, 0.32f, 0.58f, 0.22f, 0.05f, 0.0f, 0.0f);
        }

        public void Paint(Graphics graphics, float width, float size, Brush tendrilBrush, Brush bloomBrush, Color bloomPenColor)
        {
            if (IsBlossom)
            {
                PointF bloomBase = GetPointBloom(1.0f - 0.4f * Param.BlssmSize, size);
                PointF bloomCenter = GetPointBloom(1.0f, size);
                PointF bloomMiddleVector = new PointF(2.0f * (bloomCenter.X - bloomBase.X), 2.0f * (bloomCenter.Y - bloomBase.Y));
                PointF bloomTransverseFixed = new PointF(-bloomMiddleVector.Y, bloomMiddleVector.X);
                PointF bloomTransverse = new PointF(-bloomMiddleVector.Y * Param.Bloom, bloomMiddleVector.X * Param.Bloom);
                PointF bloomEdge = new PointF(bloomBase.X + (0.7f + 0.3f * Param.Bloom) * bloomMiddleVector.X + 0.9f * bloomTransverse.X, bloomBase.Y + (0.7f + 0.3f * Param.Bloom) * bloomMiddleVector.Y + 0.9f * bloomTransverse.Y);
                PointF bloomLow = new PointF(bloomCenter.X + 0.9f * bloomTransverseFixed.X - 0.9f * bloomMiddleVector.X, bloomCenter.Y + 0.9f * bloomTransverseFixed.Y - 0.9f * bloomMiddleVector.Y);
                float s = 0.8f - 1.8f * Param.Bloom * (1.0f - Param.Bloom);
                PointF bloomSideCenter = new PointF(0.9f * bloomCenter.X + (0.6f - Param.Bloom * 0.5f) * bloomTransverseFixed.X + 0.1f * bloomBase.X + s * bloomMiddleVector.X, 0.9f * bloomCenter.Y + (0.6f - Param.Bloom * 0.5f) * bloomTransverseFixed.Y + 0.1f * bloomBase.Y + s * bloomMiddleVector.Y);
                PointF bloomEdge2 = new PointF(bloomBase.X + (0.7f + 0.3f * Param.Bloom) * bloomMiddleVector.X - 0.9f * bloomTransverse.X, bloomBase.Y + (0.7f + 0.3f * Param.Bloom) * bloomMiddleVector.Y - 0.9f * bloomTransverse.Y);
                PointF bloomLow2 = new PointF(bloomCenter.X - 0.9f * bloomTransverseFixed.X - 0.9f * bloomMiddleVector.X, bloomCenter.Y - 0.9f * bloomTransverseFixed.Y - 0.9f * bloomMiddleVector.Y);
                PointF bloomSideCenter2 = new PointF(0.9f * bloomCenter.X - (0.6f - Param.Bloom * 0.5f) * bloomTransverseFixed.X + 0.1f * bloomBase.X + s * bloomMiddleVector.X, 0.9f * bloomCenter.Y - (0.6f - Param.Bloom * 0.5f) * bloomTransverseFixed.Y + 0.1f * bloomBase.Y + s * bloomMiddleVector.Y);

                using (GraphicsPath path = new GraphicsPath())
                {
                    path.AddBezier(
                        bloomBase.X, bloomBase.Y, bloomLow.X, bloomLow.Y,
                        bloomSideCenter.X, bloomSideCenter.Y, bloomEdge.X, bloomEdge.Y);
                    path.AddBezier(
                        bloomEdge.X, bloomEdge.Y,
                        bloomEdge.X + Param.Bloom * (0.4f * bloomMiddleVector.X - 0.4f * bloomTransverse.X), bloomEdge.Y + Param.Bloom * (0.4f * bloomMiddleVector.Y - 0.4f * bloomTransverse.Y),
                        bloomCenter.X + (0.2f + 0.2f * Param.Bloom) * bloomMiddleVector.X + 0.3f * bloomTransverse.X, bloomCenter.Y + (0.2f + 0.2f * Param.Bloom) * bloomMiddleVector.Y + 0.3f * bloomTransverse.Y,
                        bloomCenter.X + (0.2f + 0.3f * Param.Bloom) * bloomMiddleVector.X - 0.1f * bloomTransverse.X, bloomCenter.Y + (0.2f + 0.3f * Param.Bloom) * bloomMiddleVector.Y - 0.1f * bloomTransverse.Y);
                    path.AddBezier(
                        bloomCenter.X + (0.2f + 0.2f * Param.Bloom) * bloomMiddleVector.X + 0.2f * bloomTransverse.X, bloomCenter.Y + (0.2f + 0.2f * Param.Bloom) * bloomMiddleVector.Y + 0.2f * bloomTransverse.Y,
                        bloomCenter.X + (0.2f + 0.1f * Param.Bloom) * bloomMiddleVector.X - 0.3f * bloomTransverse.X, bloomCenter.Y + (0.2f + 0.1f * Param.Bloom) * bloomMiddleVector.Y - 0.3f * bloomTransverse.Y,
                        bloomEdge2.X + Param.Bloom * (0.4f * bloomMiddleVector.X + 0.4f * bloomTransverse.X), bloomEdge2.Y + Param.Bloom * (0.4f * bloomMiddleVector.Y + 0.4f * bloomTransverse.Y),
                        bloomEdge2.X, bloomEdge2.Y);
                    path.AddBezier(
                        bloomEdge2.X, bloomEdge2.Y, bloomSideCenter2.X, bloomSideCenter2.Y,
                        bloomLow2.X, bloomLow2.Y, bloomBase.X, bloomBase.Y);
                    graphics.FillPath(bloomBrush, path);
                }

                const int numPoints = 30;
                PointF[] points = new PointF[numPoints];
                for (int i = 0; i < numPoints / 2; i++)
                {
                    float t = i / (float)(numPoints / 2 - 1);
                    PointF p = GetPointBloom(t, size);
                    PointF r = GetRibBloom(t);
                    float w = 0.5f * width * GetRelativeWidthBloom(t);
                    points[i] = new PointF(p.X - w * r.X, p.Y - w * r.Y);
                    points[numPoints - 1 - i] = new PointF(p.X + w * r.X, p.Y + w * r.Y);
                }
                graphics.FillClosedCurve(tendrilBrush, points);

                //graphics.FillEllipse(Brushes.Blue, bloomBase.X, bloomBase.Y, 0.03f, 0.03f);
                //graphics.FillEllipse(Brushes.Red, bloomLow.X, bloomLow.Y, 0.03f, 0.03f);
                //graphics.FillEllipse(Brushes.Green, bloomSideCenter.X, bloomSideCenter.Y, 0.03f, 0.03f);
                //graphics.FillEllipse(Brushes.Violet, bloomEdge.X, bloomEdge.Y, 0.03f, 0.03f);
                Pen pen = new Pen(bloomPenColor, size * 0.1f * Param.BlssmWdth);
                graphics.DrawBezier(pen,
                    bloomBase.X, bloomBase.Y, bloomLow.X, bloomLow.Y,
                    bloomSideCenter.X, bloomSideCenter.Y, bloomEdge.X, bloomEdge.Y);
                graphics.DrawBezier(pen,
                    bloomEdge.X, bloomEdge.Y,
                    bloomEdge.X + Param.Bloom * (0.4f * bloomMiddleVector.X - 0.4f * bloomTransverse.X), bloomEdge.Y + Param.Bloom * (0.4f * bloomMiddleVector.Y - 0.4f * bloomTransverse.Y),
                    bloomCenter.X + (0.2f + 0.2f * Param.Bloom) * bloomMiddleVector.X + 0.3f * bloomTransverse.X, bloomCenter.Y + (0.2f + 0.2f * Param.Bloom) * bloomMiddleVector.Y + 0.3f * bloomTransverse.Y,
                    bloomCenter.X + (0.2f + 0.3f * Param.Bloom) * bloomMiddleVector.X - 0.1f * bloomTransverse.X, bloomCenter.Y + (0.2f + 0.3f * Param.Bloom) * bloomMiddleVector.Y - 0.1f * bloomTransverse.Y);
                graphics.DrawBezier(pen,
                    bloomCenter.X + (0.2f + 0.2f * Param.Bloom) * bloomMiddleVector.X + 0.2f * bloomTransverse.X, bloomCenter.Y + (0.2f + 0.2f * Param.Bloom) * bloomMiddleVector.Y + 0.2f * bloomTransverse.Y,
                    bloomCenter.X + (0.2f + 0.1f * Param.Bloom) * bloomMiddleVector.X - 0.3f * bloomTransverse.X, bloomCenter.Y + (0.2f + 0.1f * Param.Bloom) * bloomMiddleVector.Y - 0.3f * bloomTransverse.Y,
                    bloomEdge2.X + Param.Bloom * (0.4f * bloomMiddleVector.X + 0.4f * bloomTransverse.X), bloomEdge2.Y + Param.Bloom * (0.4f * bloomMiddleVector.Y + 0.4f * bloomTransverse.Y),
                    bloomEdge2.X, bloomEdge2.Y);
                graphics.DrawBezier(pen,
                    bloomEdge2.X, bloomEdge2.Y, bloomSideCenter2.X, bloomSideCenter2.Y,
                    bloomLow2.X, bloomLow2.Y, bloomBase.X, bloomBase.Y);
            }
            else if (IsLeaf)
            {
                float sgn = (CurlsLeft?-1.0f:1.0f);
                Matrix mat = graphics.Transform;
                graphics.RotateTransform(sgn*40.0f);
                float sz = 0.001f+2.0f*size * Param.LeavsSize;
                graphics.ScaleTransform(-sz, -sgn * sz);
                graphics.FillPath(tendrilBrush, leafPath);
                graphics.Transform = mat;
            }
            else
            {
                const int numPoints = 80;
                PointF[] points = new PointF[numPoints];
                for (int i = 0; i < numPoints / 2; i++)
                {
                    float t = (float)Math.Sqrt((i / (0.5 * numPoints - 1.0))); // more points toward the tip
                    PointF p = GetPoint(t, size);
                    PointF r = GetRib(t);
                    float w = 0.5f * width * GetRelativeWidth(t);
                    points[i] = new PointF(p.X - w * r.X, p.Y - w * r.Y);
                    points[numPoints - 1 - i] = new PointF(p.X + w * r.X, p.Y + w * r.Y);
                }
                graphics.FillClosedCurve(tendrilBrush, points);
            }

            if (!IsLeaf)
            {
                Matrix m = graphics.Transform;
                foreach (Tendril t in Children)
                {
                    PointF p;
                    float a;
                    float w;
                    if (IsBlossom)
                    {
                        p = GetPointBloom(t.Position, size);
                        a = GetAngleBloom(t.Position);
                        w = GetRelativeWidthBloom(t.Position);
                    }
                    else
                    {
                        p = GetPoint(t.Position, size);
                        a = GetAngle(t.Position);
                        w = GetRelativeWidth(t.Position);
                    }

                    graphics.TranslateTransform(p.X, p.Y);
                    graphics.RotateTransform(a - t.Angle);
                    t.Paint(graphics,
                        width * Param.StemScale * w,
                        (float)(size * Param.SizeScale * Math.Pow(1.0 - t.Position, Param.SizeDrop)),
                        tendrilBrush, bloomBrush, bloomPenColor);
                    graphics.Transform = m;
                }
            }
        }

        public PointF GetPoint(float position, float size)
        {
            double t = 2.0 * Math.PI * Math.Pow(position, Param.SpiralTip) * (Param.SpiralWind+SpiralWindOffset);
            double y = -(1.0 - position) * Math.Sin(t);
            double x = 1.0 - (1.0 - position)* Math.Cos(t);
            if (CurlsLeft)
            {
                y = -y;
            }
            return new PointF((float)(size*x), (float)(size*y));
        }

        public float GetAngle(float position)
        {
            double t = 2.0 * Math.PI * Math.Pow(position, Param.SpiralTip) * (Param.SpiralWind+SpiralWindOffset);
            double td = 2.0 * Math.PI * Param.SpiralTip * Math.Pow(position, Param.SpiralTip - 1.0) * Param.SpiralWind;
            double y = -(1.0 - position) * Math.Cos(t) * td + Math.Sin(t);
            double x = (1.0 - position) * Math.Sin(t) * td + Math.Cos(t);
            if (CurlsLeft)
            {
                y = -y;
            }
            return (float)(180.0/Math.PI*Math.Atan2(y,x));
        }

        public PointF GetRib(float position)
        {
            double t = 2.0 * Math.PI * Math.Pow(position, Param.SpiralTip) * (Param.SpiralWind + SpiralWindOffset);
            double td = 2.0 * Math.PI * Param.SpiralTip * Math.Pow(position, Param.SpiralTip - 1.0) * Param.SpiralWind;
            double y = -(1.0 - position) * Math.Cos(t) * td + Math.Sin(t);
            double x = (1.0 - position) * Math.Sin(t) * td + Math.Cos(t);
            if (CurlsLeft)
            {
                y = -y;
            }
            double len = Math.Sqrt(x * x + y * y);
            return new PointF((float)(-y/len),(float)(x/len));
        }

        public float GetRelativeWidth(float position)
        {
            return (float)Math.Pow(1.0 - position, Param.Bulge);
        }

        public PointF GetPointBloom(float position, float size)
        {
            double ip = 1.0 - position;
            double y = 3.0 * ip * position * position * 0.3 + Param.BlssmSide * position * position * position * 0.5;
            double x = 3.0 * ip * ip * position + 3.0*ip*position*position*0.5 + position*position*position;
            if (CurlsLeft)
            {
                y = -y;
            }
            return new PointF((float)(size * x), (float)(size * y));
        }

        public float GetAngleBloom(float position)
        {
            double ip = 1.0 - position;
            double y = 3.0 * (ip * 2.0 * position - position * position) * 0.3 + Param.BlssmSide * 3.0 * position * position * 0.5;
            double x = 3.0 * (-2.0*ip * position + ip * ip) + 3.0 * (ip * 2.0 * position - position * position) * 0.5 + 3.0 * position * position;
            if (CurlsLeft)
            {
                y = -y;
            }
            return (float)(180.0 / Math.PI * Math.Atan2(y, x));
        }

        public PointF GetRibBloom(float position)
        {
            double ip = 1.0 - position;
            double y = 3.0 * (ip * 2.0 * position - position * position) * 0.3 + Param.BlssmSide * 3.0 * position * position * 0.5;
            double x = 3.0 * (-2.0 * ip * position + ip * ip) + 3.0 * (ip * 2.0 * position - position * position) * 0.5 + 3.0 * position * position;
            if (CurlsLeft)
            {
                y = -y;
            }
            double len = Math.Sqrt(x * x + y * y);
            return new PointF((float)(-y / len), (float)(x / len));
        }

        public float GetRelativeWidthBloom(float position)
        {
            return (float)Math.Pow(1.0 - position, Param.Bulge);
        }
    }
}